
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.28.2',
    'branch_nick': 'release-1.28',
    'revision_id': 'b70cc31becbd674ed7140759f7db0d73addea694',
    'revno': 11515
}
